#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <assert.h>

#include <GL/gl.h>
#include <png.h>

#include "font.h"
#include "panic.h"
#include "dict.h"
#include "font_render.h"

struct font_render_info {
  	struct glyph_info glyph_table[256];
  	int texture_width;
	int texture_height;
	GLuint texture_id;
	GLuint display_list_base;
};

struct image {
	int width;
	int height;
	unsigned *bits;
};

static struct image *
img_make_from_png(const char *png_filename)
{
	FILE *fp;
	struct image *img;
	png_structp png_ptr;
	png_bytep *rows;
	int i;

	if ((fp = fopen(png_filename, "rb")) == NULL)
		panic("fopen");

	if ((png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL,
	  NULL, NULL)) == NULL)
		panic("png_create_read_struct");

	png_infop info_ptr;

	if ((info_ptr = png_create_info_struct(png_ptr)) == NULL)
		panic("png_create_info_struct");

	if (setjmp(png_jmpbuf(png_ptr)))
		panic("png error");

	png_init_io(png_ptr, fp);

	png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);

	if (png_get_color_type(png_ptr, info_ptr) != PNG_COLOR_TYPE_RGBA)
		panic("invalid color type");

	img = malloc(sizeof *img);

	img->width = png_get_image_width(png_ptr, info_ptr);
	img->height = png_get_image_height(png_ptr, info_ptr);

	img->bits = malloc(img->width*img->height*sizeof *img->bits);

	rows = png_get_rows(png_ptr, info_ptr);

	for (i = 0; i < img->height; i++)
		memcpy(&img->bits[i*img->width], rows[i],
		  img->width*sizeof *img->bits);

	png_destroy_read_struct(&png_ptr, &info_ptr, NULL);

	fclose(fp);

	return img;
}

static void
img_free(struct image *img)
{
	free(img->bits);
	free(img);
}

struct font_render_info *
font_render_info_make(const struct font_info *info)
{
	struct font_render_info *fri;
	struct image *img;
	struct glyph_info *g;
	GLuint texture_id;
	
	img = img_make_from_png(info->texture_filename);

	glEnable(GL_TEXTURE_2D);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glGenTextures(1, &texture_id);

	if (glGetError() != GL_NO_ERROR)
		panic("glGenTextures failed");

printf("texture_id: %d\n", texture_id);

	glBindTexture(GL_TEXTURE_2D, texture_id);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, img->width, img->height,
	  0, GL_RGBA, GL_UNSIGNED_BYTE, img->bits);

	glDisable(GL_TEXTURE_2D);

	fri = malloc(sizeof *fri);

	fri->texture_id = texture_id;

	/* create display lists */

	float ds = 1.f/img->width;
	float dt = 1.f/img->height;

	fri->display_list_base = glGenLists(128);

	for (g = info->glyphs; g != &info->glyphs[info->num_glyphs]; g++) {
		float s0, s1, t0, t1;

		assert(g->code >= 0 && g->code < 128);

		glNewList(fri->display_list_base + g->code, GL_COMPILE);

		glBindTexture(GL_TEXTURE_2D, fri->texture_id);

		glBegin(GL_QUADS);

		s0 = ds*g->texture_x;
		s1 = ds*(g->texture_x + (g->transposed ? g->height : g->width)); 
		t0 = dt*g->texture_y;
		t1 = dt*(g->texture_y + (g->transposed ? g->width : g->height));

		glTexCoord2f(s0, t0);
		glVertex2f(g->left, -g->top);

		if (g->transposed)
			glTexCoord2f(s0, t1);
		else
			glTexCoord2f(s1, t0);
		glVertex2f(g->left + g->width, -g->top);

		glTexCoord2f(s1, t1);
		glVertex2f(g->left + g->width, g->height - g->top);

		if (g->transposed)
			glTexCoord2f(s1, t0);
		else
			glTexCoord2f(s0, t1);
		glVertex2f(g->left, g->height - g->top);

		glEnd();

		glTranslatef(g->advance_x, 0, 0);

		glEndList();
	}

	img_free(img);

	return fri;
}

void
render_string(const struct font_render_info *fri, const char *str, int x, int y)
{
	GLint viewport[4];
	const char *p;

	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glEnable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);
	glDisable(GL_CULL_FACE);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glBindTexture(GL_TEXTURE_2D, fri->texture_id);

	glGetIntegerv(GL_VIEWPORT, viewport);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0.f, viewport[2], viewport[3], 0.f, -1.f, 1.f);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glColor4f(1.f, 1.f, 1.f, 1.f);

	glTranslated(x, y, 0);

	for (p = str; *p; p++)
		glCallList(fri->display_list_base + *p);

	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glPopAttrib();
}
