/* render.c -- part of vulcan
 *
 * This program is copyright (C) 2006 Mauro Persano, and is free
 * software which is freely distributable under the terms of the
 * GNU public license, included as the file COPYING in this
 * distribution.  It is NOT public domain software, and any
 * redistribution not permitted by the GNU General Public License is
 * expressly forbidden without prior written permission from
 * the author.
 *
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include "panic.h"
#include "game.h"
#include "vector.h"
#include "matrix.h"
#include "mesh.h"
#include "graphics.h"
#include "model.h"

#define MAX_ATTACK_BOARDS (NUM_MAIN_BOARDS*8)
#define MAX_BOARDS (NUM_MAIN_BOARDS + MAX_ATTACK_BOARDS)

extern struct mesh *unit_cylinder;
extern struct model piece_models[NUM_PIECES];

struct rgb {
	float r, g, b;
};

struct material {
	struct rgb ambient;
	struct rgb diffuse;
	struct rgb specular;
	int shine;
};

static const float board_alpha = .6f;

/* colors for black square for each selection state */
static struct rgb black_rgb[4] =
  { { .2f, .2f, .2f },		/* neutral */
    { .1f, .6f, .1f },		/* selected square */
    { .6f, .2f, 1.f },		/* attacked square */
    { .5f, .5f, .5f } };	/* highlighted square */

/* colors for white square for each selection state */
static struct rgb white_rgb[4] =
  { { .8f, .8f, .8f },		/* neutral */
    { .4f, 1.f, .4f },		/* selected square */
    { .6f, .2f, 1.f },		/* attacked square */
    { 1.f, 1.f, 1.f } };	/* highlighted square */

/* materials for white piece when unselected, selected or attacked */
static const struct material white_piece_material[4] =
  { { { .2f, .2f, .2f }, { .6f, .6f, .2f }, { .9f, .9f, .9f }, 8 },
    { { .2f, .2f, .2f }, { .4f, 1.f, .4f }, { .9f, .9f, .9f }, 8 },
    { { .2f, .2f, .2f }, { .6f, .2f, 1.f }, { .9f, .9f, .9f }, 8 },
    { { .6f, .6f, .6f }, { 1.f, 1.f, .6f }, { 1.f, 1.f, 1.f }, 2 } };

/* materials for black piece when unselected, selected or attacked */
static const struct material black_piece_material[4] = 
  { { { .2f, .2f, .2f }, { .9f, .2f, .2f }, { .9f, .9f, .9f }, 8 },
    { { .2f, .2f, .2f }, { .4f, 1.f, .4f }, { .9f, .9f, .9f }, 8 },
    { { .2f, .2f, .2f }, { .6f, .2f, 1.f }, { .9f, .9f, .9f }, 8 },
    { { .4f, .4f, .4f }, { 1.f, .4f, .4f }, { 1.f, 1.f, 1.f }, 2 } };

/* display lists */
static GLuint piece_disp_lists[NUM_PIECES];
static GLuint cylinder_disp_list;


void
mat_to_opengl(float *r, const struct matrix *m)
{
	r[0] = m->m11;  r[4] = m->m12;  r[8] = m->m13; r[12] = m->m14;
	r[1] = m->m21;  r[5] = m->m22;  r[9] = m->m23; r[13] = m->m24;
	r[2] = m->m31;  r[6] = m->m32; r[10] = m->m33; r[14] = m->m34;
	r[3] = 0.f;     r[7] = 0.f;    r[11] = 0.f;    r[15] = 1.f;
}

void
set_opengl_matrix(const struct matrix *m)
{
	static float fm[16];

	glLoadIdentity();
	mat_to_opengl(fm, m);
        glMultMatrixf(fm);
}


void
set_opengl_material(const struct material *mat)
{
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, (float *)&mat->diffuse);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, (float *)&mat->specular);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, mat->shine);
}


static void
render_mesh(const struct mesh *mesh)
{
	struct polygon *poly, *end;
	struct vertex *mesh_vtx, *v0, *v1, *v2;
	int *vtx_index;
	int i;

	glBegin(GL_TRIANGLES);

	end = &mesh->poly[mesh->npoly];
	mesh_vtx = mesh->vtx;

	for (poly = mesh->poly; poly != end; poly++) {
		vtx_index = poly->vtx_index;
		v0 = &mesh_vtx[vtx_index[0]];

		for (i = 1; i < poly->nvtx - 1; i++) {
			v1 = &mesh_vtx[vtx_index[i]];
			v2 = &mesh_vtx[vtx_index[i + 1]];

			glNormal3f(v0->normal.x, v0->normal.y, v0->normal.z);
			glVertex3f(v0->pos.x, v0->pos.y, v0->pos.z);

			glNormal3f(v1->normal.x, v1->normal.y, v1->normal.z);
			glVertex3f(v1->pos.x, v1->pos.y, v1->pos.z);

			glNormal3f(v2->normal.x, v2->normal.y, v2->normal.z);
			glVertex3f(v2->pos.x, v2->pos.y, v2->pos.z);
		}
	}

	glEnd();
}


static void
init_display_list_from_mesh(GLuint display_list, struct mesh *mesh)
{
	glNewList(display_list, GL_COMPILE);

	if (glGetError() != GL_NO_ERROR)
		panic("glNewList failed");

	render_mesh(mesh);

	glEndList();
}


static void
render_display_list(GLuint display_list, const struct matrix *m)
{
        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();

	set_opengl_matrix(m);

	glCallList(display_list);

	glPopMatrix();
}


static void
render_board_pieces(const struct matrix *mv, const struct vector *pos,
  const unsigned char *state, const unsigned char *flags, int size)
{
	int i, j, s;
	float x, y, z, start_x, start_y;
	unsigned const char *pf, *ps;

	start_x = pos->x - .5f*SQUARE_SIZE*(size - 1);
	start_y = pos->y - .5f*SQUARE_SIZE*(size - 1);
	z = pos->z;

	glEnable(GL_LIGHTING);
        glEnable(GL_LIGHT0);
	glShadeModel(GL_SMOOTH);
	/* glEnable(GL_CULL_FACE); */
	glDisable(GL_BLEND);

	x = start_x;
	pf = flags;
	ps = state;

	for (i = 0; i < size; i++) {
		y = start_y;

		for (j = 0; j < size; j++) {
			s = *ps++;

			if (s != EMPTY) {
				struct matrix m, t;
				int p;

				mat_make_translation(&t, x, y, z);
				mat_mul_copy(&m, mv, &t);

				if (s & BLACK_FLAG) {

					/* if it's a black piece, rotate it
					   180 degrees around the z axis */

					struct matrix r;

					mat_make_rotation_around_z(&r, M_PI);
					mat_mul_copy(&m, &m, &r);

					set_opengl_material(
					    &black_piece_material[*pf]);
				} else {
					set_opengl_material(
					    &white_piece_material[*pf]);
				}

				p = (s & PIECE_MASK) - PAWN;

				mat_mul_copy(&m, &m,
				  &piece_models[p].transform);

				render_display_list(piece_disp_lists[p], &m);
			}

			y += SQUARE_SIZE;

			++pf;
		}

		ps += BCOLS - size;
		pf += BCOLS - size;
		x += SQUARE_SIZE;
	}
}


static void
render_board_squares(const struct matrix *mv, const struct vector *pos,
  const unsigned char *state, const unsigned char *flags, int size)
{
	int i, j;
	float x, y, z, start_x, start_y;
	struct rgb *color;
	unsigned const char *pf;
	int s;

	start_x = pos->x - .5f*SQUARE_SIZE*size;
	start_y = pos->y - .5f*SQUARE_SIZE*size;
	z = pos->z;

	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);
	glDisable(GL_CULL_FACE);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();

	set_opengl_matrix(mv);

	glBegin(GL_QUADS);

	x = start_x;
	pf = flags;

	for (i = 0; i < size; i++) {
		y = start_y;

		for (j = 0; j < size; j++) {
			s = *pf++;

			color = &(((i^j)&1) ? black_rgb : white_rgb)[s];

			glColor4f(color->r, color->g, color->b, board_alpha);

			glVertex3f(x, y, z);
			glVertex3f(x + SQUARE_SIZE, y, z);
			glVertex3f(x + SQUARE_SIZE, y + SQUARE_SIZE, z);
			glVertex3f(x, y + SQUARE_SIZE, z);

			y += SQUARE_SIZE;
		}

		pf += BCOLS - size;
		x += SQUARE_SIZE;
	}

	glEnd();

	glPopMatrix();
}


static void
render_board(const struct matrix *mv, const struct vector *pos,
  const unsigned char *state, const unsigned char *flags, int size)
{
	render_board_pieces(mv, pos, state, flags, size);
	render_board_squares(mv, pos, state, flags, size);
}


/*
 * render_pins --
 *	Render the pins that connect main boards to attack boards.
 */
static void
render_pins(const struct game_state *state, const struct matrix *mv)
{
	struct matrix m, s, t;
	struct vector pos;
	int i, j;

	/* render cylinders */

	glEnable(GL_NORMALIZE);
	glEnable(GL_LIGHTING);
        glEnable(GL_LIGHT0);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_CULL_FACE);
	glDisable(GL_BLEND);

	for (i = 0; i < NUM_MAIN_BOARDS; i++) {
		for (j = 0; j < 8; j++) {
			if (ATTACK_BOARD_IS_ACTIVE(state, i, j)) {
				get_attack_board_xyz(&pos, i, j);

				mat_make_scale_xyz(&s, 3.f, 3.f,
				  ATTACK_BOARD_DISTANCE);

				if (ATTACK_BOARD_IS_ABOVE(j))
					pos.z -= ATTACK_BOARD_DISTANCE;

				mat_make_translation_from_vec(&t, &pos);
				mat_mul_copy(&m, &t, &s);
				mat_mul_copy(&m, mv, &m);

				set_opengl_material(
				  ATTACK_BOARD_IS_BLACK(state, i, j) ?
				    &black_piece_material[0] :
				    &white_piece_material[0]);

				render_display_list(cylinder_disp_list, &m);
			}
		}
	}

	glDisable(GL_NORMALIZE);
}


/*
 * render_boards --
 *	Render main boards and attack boards for the given game state, in
 *	back-to-front order.
 */
static void
render_boards(const struct game_state *state,
  const struct selected_squares *flags, const struct matrix *mv)
{
	static struct board_to_render {
		struct vector orig_pos;		/* untransformed position of
						 * center of the board */
		struct vector pos;		/* transformed position */
		int size;			/* in squares */
		const unsigned char *state;
		const unsigned char *flags;
	} boards[MAX_BOARDS];
	int render_order[MAX_BOARDS];
	int num_boards;
	struct board_to_render *p;
	struct position pos;
	int i, j;

	/* sort boards back to front (for transparency) */

	p = boards;
	num_boards = 0;

	for (i = 0; i < NUM_MAIN_BOARDS; i++) {
		/* main board */

		assert(num_boards < MAX_BOARDS);

		p->size = MAIN_BOARD_SIZE;
		get_main_board_xyz(&p->orig_pos, i);
		mat_transform_copy(&p->pos, mv, &p->orig_pos);

		get_main_board_position(&pos, i);
		p->state = &state->board[pos.level][pos.square];
		p->flags = &flags->selection[pos.level][pos.square];

		++p;
		++num_boards;

		/* attack boards */

		for (j = 0; j < 8; j++) {
			if (ATTACK_BOARD_IS_ACTIVE(state, i, j)) {
				assert(num_boards < MAX_BOARDS);

				p->size = ATTACK_BOARD_SIZE;
				get_attack_board_xyz(&p->orig_pos, i, j);
				mat_transform_copy(&p->pos, mv, &p->orig_pos);

				get_attack_board_position(&pos, i, j);
				p->state = &state->board[pos.level][pos.square];
				p->flags = &flags->selection[pos.level]
				  [pos.square];

				++p;
				++num_boards;
			}
		}
	}

	/* bubble sort is enough for N this small */

	for (i = 0; i < num_boards; i++)
		render_order[i] = i;

	for (i = 0; i < num_boards; i++) {
		for (j = num_boards - 1; j > i; j--) {
			if (boards[render_order[j - 1]].pos.z >
			  boards[render_order[j]].pos.z) {
				int t = render_order[j - 1];
				render_order[j - 1] = render_order[j];
				render_order[j] = t;
			}
		}
	}

	/* render boards */

	for (i = 0; i < num_boards; i++) {
		p = &boards[render_order[i]];
		render_board(mv, &p->orig_pos, p->state, p->flags, p->size);
	}
}


/*
 * render --
 *	Render a game state.
 */
void
render(const struct game_state *s, const struct selected_squares *f,
  const struct matrix *mv)
{
	render_pins(s, mv);
	render_boards(s, f, mv);
}


/*
 * init_render --
 *	Create display lists for pieces and cylinder.
 */
void
init_render(void)
{
	int i;
	GLuint base;

	if ((base = glGenLists(NUM_PIECES + 1)) == 0)
		panic("glGenLists failed");

	for (i = 0; i < NUM_PIECES; i++) {
		piece_disp_lists[i] = base + i;
		init_display_list_from_mesh(piece_disp_lists[i],
		  piece_models[i].mesh);
	}

	cylinder_disp_list = base + NUM_PIECES;
	init_display_list_from_mesh(cylinder_disp_list, unit_cylinder);
}
