/* vulcan.c -- part of vulcan
 *
 * This program is copyright (C) 2006 Mauro Persano, and is free
 * software which is freely distributable under the terms of the
 * GNU public license, included as the file COPYING in this
 * distribution.  It is NOT public domain software, and any
 * redistribution not permitted by the GNU General Public License is
 * expressly forbidden without prior written permission from
 * the author.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <assert.h>
#include "version.h"
#include "panic.h"
#include "list.h"
#include "dict.h"
#include "render.h"
#include "vector.h"
#include "matrix.h"
#include "mesh.h"
#include "bsptree.h"
#include "model.h"
#include "vrml.h"
#include "list.h"
#include "cylinder.h"
#include "pick.h"
#include "move.h"
#include "game.h"
#include "engine.h"
#include "font.h"
#include "font_render.h"
#include "vulcan-ui.h"

#define FONT_FILE "data/font/VeraSe.fontdef"
#define FONT_NAME "VeraSe"

extern int send_worker_request(struct game_context *ctx);
extern void swap_buffers(struct game_context *ctx);
extern unsigned long msecs(void);

enum {
	DBL_CLICK_INTERVAL = 800,	/* in msecs */
	CHAR_HEIGHT = 30,
};

struct model piece_models[NUM_PIECES];
struct mesh *unit_cylinder;

static struct font_render_info *font;

static int cur_width = -1;
static int cur_height = -1;


/*----------------------------------------------------------------------------
 *
 *	vrml mesh loading
 *
 */

static void
vrml_mesh_init_vtx(struct mesh *m, const struct list *points)
{
	struct list_node *ln;
	struct vertex *p;

	m->nvtx = points->length;
	p = m->vtx = malloc(m->nvtx * sizeof *m->vtx);

	for (ln = points->first; ln; ln = ln->next) {
		struct vrml_vector *v = (struct vrml_vector *)ln->data;
		vec_set(&p->pos, v->x, v->y, v->z);
		++p;
	}
}

static void
vrml_mesh_init_polys(struct mesh *m, const struct list *ifaces)
{
	struct list_node *ln;
	struct polygon *p, *end;
	int npoly;
	long idx;

	/* count polygons */

	npoly = 0;

	for (ln = ifaces->first; ln; ln = ln->next) {
		idx = (long)ln->data;

		if (idx == -1 || ln->next == NULL)
			++npoly;
	}

	/* initialize polygons */

	m->npoly = npoly;
	p = m->poly = malloc(m->npoly * sizeof *m->poly);

	end = &m->poly[m->npoly];
	ln = ifaces->first;

	for (p = m->poly; p != end; p++) {
		p->nvtx = 0;

		while (ln != NULL) {
			idx = (long)ln->data;
			ln = ln->next;

			if (idx == -1)
				break;

			assert(p->nvtx != MAX_POLY_VTX);

			p->vtx_index[p->nvtx++] = idx;
		}

		assert(p->nvtx == 3);
	}
}

static struct mesh *
vrml_mesh_make(struct vrml_node_coordinate3 *coord_node,
  struct vrml_node_indexed_face_set *iface_node)
{
	struct mesh *m;

	m = mesh_make();

	vrml_mesh_init_vtx(m, coord_node->points);
	vrml_mesh_init_polys(m, iface_node->indexes);

	return m;
}

struct vrml_context {
	struct vrml_node_coordinate3 *cur_coords_node;
};

static void
vrml_visit_node_list(struct vrml_context *ctx, const struct list *node_list,
  struct list *mesh_list)
{
	struct vrml_context cur_ctx = *ctx;
	struct list_node *ln;

	for (ln = node_list->first; ln; ln = ln->next) {
		union vrml_node *node = (union vrml_node *)ln->data;

		switch (node->common.type) {
			case VRML_NODE_SEPARATOR:
				vrml_visit_node_list(&cur_ctx,
				  node->separator.children, mesh_list);
				break;

			case VRML_NODE_COORDINATE3:
				cur_ctx.cur_coords_node = &node->coordinate3;
				break;

			case VRML_NODE_INDEXED_FACE_SET:
				if (cur_ctx.cur_coords_node == NULL)
					panic("vrml missing coords node?");

				list_append(mesh_list,
				  vrml_mesh_make(cur_ctx.cur_coords_node,
				    &node->indexed_face_set));
				break;

			default:
				break;
		}
	}
}

static struct list *
vrml_extract_meshes(struct list *vrml_node_list)
{
	struct vrml_context ctx = {0};
	struct list *mesh_list;

	mesh_list = list_make();

	vrml_visit_node_list(&ctx, vrml_node_list, mesh_list);

	return mesh_list;
}

static void
load_piece(struct model *piece_model, const char *file_name)
{
	struct list *vrml_node_list;
	struct list *mesh_list;
	struct mesh *mesh;
	struct matrix s, rz, rx;

	if ((vrml_node_list = vrml_parse_file(file_name)) == NULL)
		panic("failed to load %s", file_name);

	mesh_list = vrml_extract_meshes(vrml_node_list);
	list_free(vrml_node_list, (void(*)(void *))vrml_node_free);

	assert(mesh_list->length == 1);

	mesh = (struct mesh *)mesh_list->first->data;
	list_free(mesh_list, NULL);

	mat_make_scale(&s, .095f);
	mesh_transform_all_vertices(mesh, &s);

	mesh_init_poly_normals(mesh);
	mesh_init_vertex_normals(mesh);

	piece_model->mesh = mesh;

	piece_model->bsp_tree = bsp_tree_build_from_mesh(mesh);

	mat_make_rotation_around_z(&rz, -.5*M_PI);
	mat_make_rotation_around_x(&rx, .5*M_PI);
	mat_mul_copy(&piece_model->transform, &rz, &rx);

	model_find_bounding_sphere(piece_model);
}

static void
load_models(void)
{
	int i;

	static const char *mesh_names[NUM_PIECES] = {
		"data/models/pawn-low.wrl", "data/models/rook-low.wrl",
		"data/models/knight-low.wrl", "data/models/bishop-low.wrl",
		"data/models/queen-low.wrl", "data/models/king-low.wrl" };

	for (i = 0; i < NUM_PIECES; i++) {
		load_piece(&piece_models[i], mesh_names[i]);

		printf("loaded %s: %d vtx, %d poly, %d nodes in bsp tree\n",
		  mesh_names[i],
		  piece_models[i].mesh->nvtx,
		  piece_models[i].mesh->npoly,
		  piece_models[i].bsp_tree->num_children + 1);
	}
}

static enum player_type
cur_player_type(struct game_context *ctx)
{
	return ctx->players[ctx->cur_side == BLACK_FLAG];
}


/*----------------------------------------------------------------------------
 *
 *	drawing
 *
 */

static void
set_modelview_matrix(struct game_context *ctx, struct matrix *mv)
{
	struct matrix t;

        mat_make_translation_from_vec(&t, &ctx->board_pos);
        mat_mul_copy(mv, &t, &ctx->rmatrix);
}

void
redraw(struct game_context *ctx)
{
	struct matrix mv;
	const int baseline = 22;

	glClearColor(0.f, 0.f, 0.f, 0.f);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	set_modelview_matrix(ctx, &mv);
	render(&ctx->state, &ctx->selected_squares, &mv);

	render_string(font, ctx->last_move_buf, 5, cur_height - 22);

	switch (ctx->computer_player_state) {
		case DONE_MATE_FOR_WHITE:
			render_string(font, "Mate for white!",
			  5, cur_height - (baseline + CHAR_HEIGHT + 2));
			break;

		case DONE_MATE_FOR_BLACK:
			render_string(font, "Mate for black!",
			  5, cur_height - (baseline + CHAR_HEIGHT + 2));
			break;

		default:
			render_string(font,
			  ctx->cur_side == BLACK_FLAG ?
			  "Black's turn." : "White's turn.",
			  5, cur_height - (baseline + CHAR_HEIGHT + 2));

			render_string(font,
  			  cur_player_type(ctx) == COMPUTER_PLAYER ?
			    "Computer thinking..." : "Your move?",
			  5, cur_height - (baseline + 2*(CHAR_HEIGHT + 2)));
			break;
	}

	swap_buffers(ctx);
}

void
reshape(struct game_context *ctx, int width, int height)
{
	glViewport(0, 0, width, height);

	cur_width = width;
	cur_height = height;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45., (GLdouble)width/(GLdouble)height, 1., 1000.);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

static void
init_opengl_state(void)
{
	glClearColor(0.f, 0.f, 0.f, 0.f);
	glClearDepth(1.f);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
}

/*----------------------------------------------------------------------------
 *
 *	initialization
 *
 */

static void
init_font(void)
{
	struct dict *font_dict;
	struct font_info *fi;

	font_dict = dict_make();

	fontdef_parse_file(font_dict, FONT_FILE);

	if ((fi = dict_get(font_dict, FONT_NAME)) == NULL)
		panic("font not found in fontdef file?");

	font = font_render_info_make(fi);
}

void
initialize_gl(struct game_context *ctx)
{
	mat_make_identity(&ctx->rmatrix);
	vec_set(&ctx->board_pos, 0.f, 0.f, -320.f);
	init_opengl_state();
	init_render();
	init_font();
}

void
initialize_ui(struct game_context *ctx, enum player_type white_player,
  enum player_type black_player, int max_depth)
{
	init_move_tables();
	init_engine();
	load_models();

	unit_cylinder = make_cylinder(10.f, 1.f, 1.f);

	init_game_state(&ctx->state);

	memset(&ctx->selected_squares, UNSELECTED_SQUARE,
	  sizeof ctx->selected_squares);

	ctx->computer_player_state = IDLE;
	ctx->cur_side = 0;
	ctx->left_button_down = 0;
	ctx->right_button_down = 0;
	ctx->control_key_down = 0;
	ctx->start_position_selected = 0;
	ctx->pos_highlighted = 0;
	ctx->last_clicked_x = -1;
	ctx->last_clicked_y = -1;
	ctx->last_click_msecs = 0;
	ctx->max_depth = max_depth;

	ctx->players[0] = white_player;
	ctx->players[1] = black_player;

	ctx->last_move_buf[0] = 0;

	ctx->timer_list = NULL;
}

/*----------------------------------------------------------------------------
 *
 *	event handling
 *
 */

static void
mark_selected_square(struct selected_squares *selected,
  const struct position *pos, enum selection_type t)
{
	selected->selection[pos->level][pos->square] = t;
}

static void
mark_selected_attack_board(struct selected_squares *selected,
  const struct attack_board *pos, enum selection_type t)
{
	int i, j;
	struct position first_pos;
	unsigned char *p;

	get_attack_board_position(&first_pos, pos->main_board, pos->position);

	p = &selected->selection[first_pos.level][first_pos.square];

	for (i = 0; i < ATTACK_BOARD_SIZE; i++) {
		for (j = 0; j < ATTACK_BOARD_SIZE; j++)
			*p++ = t;

		p += BCOLS - ATTACK_BOARD_SIZE;
	}
}

static void
mark_selected_squares(struct selected_squares *selected,
  const struct game_state *state, 
  int side, union move *moves, int num_moves)
{
	union move *move, *end;
	const struct position *pos;
	unsigned char s;
	enum selection_type t;
	struct position closest;

	end = &moves[num_moves];

	for (move = moves; move != end; move++) {
		switch (move->type) {
			case PIECE_MOVE:
				mark_selected_square(selected,
				  &move->piece_move.from, SELECTED_SQUARE);

				pos = &move->piece_move.to;

				s = state->board[pos->level][pos->square];

				if (s == EMPTY || (s & BLACK_FLAG) == side)
					t = SELECTED_SQUARE;
				else
					t = ATTACKED_SQUARE;

				mark_selected_square(selected, pos, t);
				break;

			case ATTACK_BOARD_MOVE:
				mark_selected_attack_board(selected,
				  &move->attack_board_move.from,
				    SELECTED_SQUARE);

				/* mark clickable destination square */

				get_closest_main_board_square_for_attack_board(
				  &closest, &move->attack_board_move.to);

				mark_selected_square(selected, &closest,
				  ATTACKED_SQUARE);
				break;

			case WHITE_KINGSIDE_CASTLING:
				selected->selection[4][11*BCOLS + 5] =
				  SELECTED_SQUARE;
				selected->selection[4][11*BCOLS + 6] =
				  ATTACKED_SQUARE;
				break;

			case WHITE_QUEENSIDE_CASTLING:
				selected->selection[4][11*BCOLS + 5] =
				  SELECTED_SQUARE;
				selected->selection[4][11*BCOLS + 2] =
				  SELECTED_SQUARE;
				break;

			case BLACK_KINGSIDE_CASTLING:
				selected->selection[0][2*BCOLS + 5] =
				  SELECTED_SQUARE;
				selected->selection[0][2*BCOLS + 6] =
				  ATTACKED_SQUARE;
				break;

			case BLACK_QUEENSIDE_CASTLING:
				selected->selection[0][2*BCOLS + 5] =
				  SELECTED_SQUARE;
				selected->selection[0][2*BCOLS + 2] =
				  SELECTED_SQUARE;
				break;

			default:
				assert(0);
		}
	}
}

static void
on_start_position_selected(struct game_context *ctx, struct position *pos)
{
	union move moves[MAX_MOVES];
	int n;

	if ((n = get_legal_moves_for_position(moves, pos, &ctx->state,
	  ctx->cur_side)) > 0) {
		mark_selected_squares(&ctx->selected_squares,
		  &ctx->state, ctx->cur_side, moves, n);

		mark_selected_square(&ctx->selected_squares, pos,
		  SELECTED_SQUARE);

		ctx->last_selected_pos = *pos;
		ctx->start_position_selected = 1;
		ctx->pos_highlighted = 0;
	}
}

void
print_move(union move *move)
{
	if (move->type == PIECE_MOVE) {
		const struct piece_move *pm = &move->piece_move;

		printf("piece: from %d,%d to %d,%d\n",
		  pm->from.level, pm->from.square,
		  pm->to.level, pm->to.square);
	} else {
		const struct attack_board_move *am = &move->attack_board_move;

		printf("attack board: from %d-%d to %d-%d\n",
		  am->from.main_board, am->from.position,
		  am->to.main_board, am->to.position);
	}
}

static void
game_do_move(struct game_context *ctx, union move *move)
{
	struct undo_move_info undo_info;

	sprintf(ctx->last_move_buf, "%d. %s %s", 1 + ctx->state.cur_ply/2,
	  (ctx->state.cur_ply & 1) ? "..." : "",
	  move_as_string(&ctx->state, move));

	do_move(&ctx->state, move, &undo_info);

	if (get_game_status(&ctx->state) != IN_PLAY) {
		ctx->computer_player_state =
		  ctx->cur_side == BLACK_FLAG ?
		    DONE_MATE_FOR_BLACK : DONE_MATE_FOR_WHITE;

		strcat(ctx->last_move_buf, "++");
	} else {
		ctx->cur_side ^= BLACK_FLAG;
		ctx->computer_player_state = IDLE;

		if (is_in_check(&ctx->state, ctx->cur_side))
			strcat(ctx->last_move_buf, "+");
	}
}

static void
on_end_position_selected(struct game_context *ctx, struct position *pos)
{
	union move moves[MAX_MOVES];
	union move *selected_move;
	struct position closest, *start;
	int i, n;

	start = &ctx->last_selected_pos;

	assert((n = get_legal_moves_for_position(moves,
	  start, &ctx->state, ctx->cur_side)) > 0);

	selected_move = NULL;

	for (i = 0; i < n && selected_move == NULL; i++) {
		switch (moves[i].type) {
			case PIECE_MOVE:
				if (SAME_POS(&moves[i].piece_move.to, pos)) {
					selected_move = &moves[i];
					break;
				}
				break;

			case ATTACK_BOARD_MOVE:
				get_closest_main_board_square_for_attack_board(
				  &closest, &moves[i].attack_board_move.to);

				if (SAME_POS(&closest, pos)) {
					selected_move = &moves[i];
					break;
				}
				break;

			case WHITE_KINGSIDE_CASTLING:
				if (start->level == 4 &&
				  start->square == 11*BCOLS+5 &&
				  pos->level == 4 &&
				  pos->square == 11*BCOLS+6) {
					selected_move = &moves[i];
					break;
				}
				break;

			case WHITE_QUEENSIDE_CASTLING:
				if (start->level == 4 &&
				  start->square == 11*BCOLS+5 &&
				  pos->level == 4 &&
				  pos->square == 11*BCOLS+2) {
					selected_move = &moves[i];
					break;
				}
				break;

			case BLACK_KINGSIDE_CASTLING:
				if (start->level == 0 &&
				  start->square == 2*BCOLS+5 &&
				  pos->level == 0 &&
				  pos->square == 2*BCOLS+6) {
					selected_move = &moves[i];
					break;
				}
				break;

			case BLACK_QUEENSIDE_CASTLING:
				if (start->level == 0 &&
				  start->square == 2*BCOLS+5 &&
				  pos->level == 0 &&
				  pos->square == 2*BCOLS+2) {
					selected_move = &moves[i];
					break;
				}
				break;
		}
	}

	if (selected_move != NULL)
		game_do_move(ctx, selected_move);

	ctx->start_position_selected = 0;
}

static void
on_position_selected(struct game_context *ctx, struct position *pos)
{
	memset(&ctx->selected_squares, UNSELECTED_SQUARE,
	  sizeof ctx->selected_squares);

	if (!ctx->start_position_selected)
		on_start_position_selected(ctx, pos);
	else
		on_end_position_selected(ctx, pos);
}

static void
on_double_click(struct game_context *ctx, int x, int y)
{
	struct position pos;
	struct matrix mv;

	if (ctx->computer_player_state != WAITING_FOR_USER_INPUT)
		return;

	memset(&ctx->selected_squares, UNSELECTED_SQUARE,
	  sizeof ctx->selected_squares);

	set_modelview_matrix(ctx, &mv);

	if (pick_square(&pos, &ctx->state, &mv, x, y))
		on_position_selected(ctx, &pos);

	redraw(ctx);
}

void
on_worker_reply(struct game_context *ctx, union move *next_move)
{
	game_do_move(ctx, next_move);
	redraw(ctx);
}

void
on_button_press(struct game_context *ctx, int button, int x, int y)
{		
	unsigned long cur_msecs;

	if (button == BTN_LEFT || button == BTN_RIGHT) {
		cur_msecs = msecs();

		if (button == BTN_LEFT && x == ctx->last_clicked_x &&
		  y == ctx->last_clicked_y &&
		  cur_msecs - ctx->last_click_msecs <= DBL_CLICK_INTERVAL) {
			ctx->last_click_msecs = 0;
			ctx->last_clicked_x = ctx->last_clicked_y = -1;
			on_double_click(ctx, x, y);
		} else {
			if (button == BTN_LEFT)
				ctx->left_button_down = 1;
			else
				ctx->right_button_down = 1;
			ctx->last_click_msecs = cur_msecs;
			ctx->last_clicked_x = x;
			ctx->last_clicked_y = y;
		}
	}
}

void
on_button_release(struct game_context *ctx, int button, int x, int y)
{
	if (button == BTN_LEFT)
		ctx->left_button_down = 0;
	else if (button == BTN_RIGHT)
		ctx->right_button_down = 0;
}

void
on_motion_notify(struct game_context *ctx, int button, int x, int y)
{
	if (ctx->left_button_down || ctx->right_button_down) {
		/* dragging */

		int dx, dy;
		struct matrix r, rx, ry;

		if (ctx->pos_highlighted) {
			mark_selected_square(&ctx->selected_squares,
			  &ctx->last_highlighted_pos, UNSELECTED_SQUARE);

			ctx->pos_highlighted = 0;
		}

		dx = x - ctx->last_clicked_x;
		dy = y - ctx->last_clicked_y;
		ctx->last_clicked_x = x;
		ctx->last_clicked_y = y;

		if (ctx->left_button_down) {
			mat_make_rotation_around_y(&ry, (float)-dx/100.f);
			mat_make_rotation_around_x(&rx, (float)dy/100.f);
			mat_mul_copy(&r, &ry, &rx);
			mat_mul_copy(&ctx->rmatrix, &r, &ctx->rmatrix);
		} else if (ctx->control_key_down) {
			ctx->board_pos.z -= (float)dy/1.5f;
		} else {
			ctx->board_pos.x -= (float)dx*ctx->board_pos.z/500.f;
			ctx->board_pos.y += (float)dy*ctx->board_pos.z/500.f;
		}

		redraw(ctx);
	} else if (ctx->computer_player_state == WAITING_FOR_USER_INPUT) {
		/* selecting */

		struct position pos;
		struct matrix mv;

		if (ctx->pos_highlighted) {
			mark_selected_square(&ctx->selected_squares,
			  &ctx->last_highlighted_pos, UNSELECTED_SQUARE);

			ctx->pos_highlighted = 0;
		}

		set_modelview_matrix(ctx, &mv);

		if (pick_square(&pos, &ctx->state, &mv, x, y)) {
			const int s = ctx->state.board[pos.level][pos.square];

			if (s != EMPTY && (s & BLACK_FLAG) == ctx->cur_side) {
				if (!(ctx->start_position_selected && ctx->selected_squares.selection[pos.level][pos.square] == SELECTED_SQUARE)) {
					ctx->pos_highlighted = 1;
					ctx->last_highlighted_pos = pos;

					mark_selected_square(
					  &ctx->selected_squares,
					  &ctx->last_highlighted_pos,
					  HIGHLIGHTED_PIECE);
				}
			}
		}

		redraw(ctx);
	 }
}

void
on_idle(struct game_context *ctx)
{
	if (ctx->computer_player_state == IDLE) {
		if (cur_player_type(ctx) == COMPUTER_PLAYER) {
			if (send_worker_request(ctx) != 0)
				panic("couldn't send worker request?");
			ctx->computer_player_state = THINKING;
		} else {
			ctx->computer_player_state = WAITING_FOR_USER_INPUT;
		}
	}
}

void
add_timer(struct game_context *ctx, unsigned long timeout_msec, int one_shot,
  int(*action)(struct game_context *, void *), void *extra)
{
	struct timer *t = malloc(sizeof *t);

	t->start_msec = msecs();
	t->timeout_msec = timeout_msec;
	t->one_shot = one_shot;
	t->action = action;
	t->extra = extra;
	t->next = ctx->timer_list;

	ctx->timer_list = t;
}
